 ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
from bpy.props import BoolProperty, FloatProperty, IntProperty

bl_info = {
        'name': "KnockOff",
        'author': "Kjartan Tysdal",
        'description': 'Easy and straight forward control over the bevel modifier. Spacebar/search "knockOff" ',
        'category': "3d View",
        'blender': (2, 7, 9),
        'version': (0, 0, 5),
        'wiki_url': 'http://www.kjartantysdal.com/scripts',
}



class VIEW3D_OT_ko_edit_bevel(bpy.types.Operator):
    """Interactive Topo"""
    bl_idname = "view3d.ko_edit_bevel"
    bl_label = "kO Edit Bevel"

    def updateTooltip(self):
    
        self.segments = str(bpy.context.object.modifiers[self.modifier_name].segments)
        self.width = bpy.context.object.modifiers[self.modifier_name].width
        self.width = "{:3.2f}".format(self.width)
        self.profile = bpy.context.object.modifiers[self.modifier_name].profile
        self.profile = "{:1.2f}".format(self.profile)
        
        self.clamp_overlap = bpy.context.object.modifiers[self.modifier_name].use_clamp_overlap
        if self.clamp_overlap:
            clampOnOff = 'ON'
        else:
            clampOnOff = 'OFF'
        
        self.loop_slide = bpy.context.object.modifiers[self.modifier_name].loop_slide
        if self.loop_slide:
            loop_slideOnOff = 'ON'
        else:
            loop_slideOnOff = 'OFF'
        
        if self.p_pressed:
            tooltip = 'Profile: '+self.profile+' , Segments: '+self.segments+' , Clamp Overlap: '+clampOnOff+' (C), Loop Slide: '+loop_slideOnOff+' (L), Profile Control: ON (P)'
        else:
            tooltip = 'Width: '+self.width+' , Segments: '+self.segments+' , Clamp Overlap: '+clampOnOff+' (C), Loop Slide: '+loop_slideOnOff+' (L), Profile Control: OFF (P)'
        
        bpy.context.area.header_text_set(tooltip)
    
        return
    
    # Only work in Edit mode
    @classmethod    
    def poll(cls, context):
        return context.area.type == 'VIEW_3D' and (bpy.context.active_object.mode == 'OBJECT' or bpy.context.active_object.mode == 'EDIT')


    def modal(self, context, event):

        if event.type == 'MOUSEMOVE':
            delta = self.first_mouse_x - event.mouse_x
            # loop through all selected objects
            for x in bpy.context.selected_objects:
                bpy.context.scene.objects.active = bpy.data.objects[x.name]
                if event.shift:
                    if self.p_pressed:
                        bpy.context.object.modifiers[self.modifier_name].profile = self.current_profile + delta * 0.0001
                    elif self.s_pressed:
                        bpy.context.object.modifiers[self.modifier_name].segments = self.current_segments + delta * 0.01
                    else:
                        bpy.context.object.modifiers[self.modifier_name].width = abs(self.current_width + delta * 0.0001)
                elif event.alt:
                    if self.p_pressed:
                        bpy.context.object.modifiers[self.modifier_name].profile = self.current_profile + delta * 0.01
                    elif self.s_pressed:
                        bpy.context.object.modifiers[self.modifier_name].segments = self.current_segments + delta * 0.1
                    else:
                        bpy.context.object.modifiers[self.modifier_name].width = abs(self.current_width + delta * 0.01)
                else:
                    if self.p_pressed:
                        bpy.context.object.modifiers[self.modifier_name].profile = self.current_profile + delta * 0.001
                    elif self.s_pressed:
                        bpy.context.object.modifiers[self.modifier_name].segments = self.current_segments + delta * 0.05
                    else:
                        bpy.context.object.modifiers[self.modifier_name].width = abs(self.current_width + delta * 0.001)
            self.updateTooltip()
    
        if event.type in {'LEFTMOUSE', 'SPACE', 'RET', 'NUMPAD_ENTER'}:
            context.area.tag_redraw()
            context.area.header_text_set()
            # fix for a very obscure bug that happens when multiple objects are selected and you are in edit mode
            bpy.context.scene.objects.active = self.active
            return {'FINISHED'}
            
        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            # Reset tooltip
            context.area.tag_redraw()
            context.area.header_text_set()
            # fix for a very obscure bug that happens when multiple objects are selected and you are in edit mode
            bpy.context.scene.objects.active = self.active
            
            bpy.context.object.modifiers[self.modifier_name].width = self.current_width
            return {'CANCELLED'}

        
        # EXTRA FUNCTIONALITY
        elif event.type == 'L':
            if event.value == 'PRESS':
                # loop through all selected objects
                for x in bpy.context.selected_objects:
                    bpy.context.scene.objects.active = bpy.data.objects[x.name]
                    bpy.context.object.modifiers[self.modifier_name].loop_slide = not bpy.context.object.modifiers[self.modifier_name].loop_slide
                    self.updateTooltip()
        elif event.type == 'C':
            if event.value == 'PRESS':
                # loop through all selected objects
                for x in bpy.context.selected_objects:
                    bpy.context.scene.objects.active = bpy.data.objects[x.name]
                    bpy.context.object.modifiers[self.modifier_name].use_clamp_overlap = not bpy.context.object.modifiers[self.modifier_name].use_clamp_overlap
                    self.updateTooltip()
                
        elif event.type == 'P':
            if event.value == 'PRESS':
                self.s_pressed = False
                self.p_pressed = not self.p_pressed
                self.current_width = bpy.context.object.modifiers[self.modifier_name].width
                self.current_profile = bpy.context.object.modifiers[self.modifier_name].profile
                self.current_segments = bpy.context.object.modifiers[self.modifier_name].segments
                self.first_mouse_x = event.mouse_x
                self.updateTooltip()
        
        elif event.type == 'S':
            if event.value == 'PRESS':
                self.p_pressed = False
                self.s_pressed = not self.s_pressed
                self.current_width = bpy.context.object.modifiers[self.modifier_name].width
                self.current_profile = bpy.context.object.modifiers[self.modifier_name].profile
                self.current_segments = bpy.context.object.modifiers[self.modifier_name].segments
                self.first_mouse_x = event.mouse_x
                self.updateTooltip()
                
        # SCROLL WHEEL CHANGE SEGMENTS
            # ------------------------------------------------------------
        elif event.type == 'WHEELUPMOUSE':
            # loop through all selected objects
            for x in bpy.context.selected_objects:
                bpy.context.scene.objects.active = bpy.data.objects[x.name]
                
                for x in bpy.context.object.modifiers:
                    if x.name == self.modifier_name: # check if bevel modifier exists already
                        bpy.context.object.modifiers[self.modifier_name].segments -= 1
                        self.updateTooltip()
        
        elif event.type == 'WHEELDOWNMOUSE':
            # loop through all selected objects
            for x in bpy.context.selected_objects:
                bpy.context.scene.objects.active = bpy.data.objects[x.name]
                
                for x in bpy.context.object.modifiers:
                    if x.name == self.modifier_name: # check if bevel modifier exists already
                        bpy.context.object.modifiers[self.modifier_name].segments += 1
                        self.updateTooltip()
        
        return {'RUNNING_MODAL'}
    
    def invoke(self, context, event):
        
        if len(bpy.context.selected_objects) == 0:
            self.report({'WARNING'}, 'Nothing is selected.')
            return {'CANCELLED'}
        
        active_check = False
        
        # check if active object is part of the selection
        for x in bpy.context.selected_objects:
            if x.name == bpy.context.scene.objects.active.name:
                active_check = True

        if active_check == False:
            bpy.context.scene.objects.active = bpy.context.selected_objects[0]
        
        self.modifier_name = 'Bevel'
        modifier_exists = False
        
        for x in bpy.context.object.modifiers:
            if x.name == self.modifier_name: # check if bevel modifier exists already
                modifier_exists = True
        
        if modifier_exists == False:
            self.report({'WARNING'}, 'There is no "Bevel" modifier on the active object.')
            return {'CANCELLED'}
        
        # Set variables
        self.width = str(bpy.context.object.modifiers[self.modifier_name].width)
        self.segments = str(bpy.context.object.modifiers[self.modifier_name].segments)
        self.profile = str(bpy.context.object.modifiers[self.modifier_name].profile)
        self.loop_slide = bpy.context.object.modifiers[self.modifier_name].loop_slide
        self.clamp_overlap = bpy.context.object.modifiers[self.modifier_name].use_clamp_overlap
        self.active = bpy.context.scene.objects.active
        self.p_pressed = False
        self.s_pressed = False
        self.first_mouse_x = event.mouse_x
        self.current_width = bpy.context.object.modifiers[self.modifier_name].width
        self.current_profile = bpy.context.object.modifiers[self.modifier_name].profile
        self.current_segments = bpy.context.object.modifiers[self.modifier_name].segments
        
        self.updateTooltip()
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}


class VIEW3D_OT_ko_clear_bevel(bpy.types.Operator):
    """Tooltip"""
    bl_idname = "view3d.ko_clear_bevel"
    bl_label = "kO Clear Bevel"
    bl_options = {'REGISTER', 'UNDO'}
    
    
    # Only work in the 3d view
    @classmethod
    def poll(cls, context):
        return context.area.type == 'VIEW_3D' and (bpy.context.active_object.mode == 'OBJECT' or bpy.context.active_object.mode == 'EDIT')

    def execute(self, context):
        # in Edit mode you simply clear sharp and bevel edges   
        if bpy.context.active_object.mode == 'EDIT':
            bpy.ops.mesh.mark_sharp(clear=True)
            bpy.ops.transform.edge_bevelweight(value=-1)
            
        # In Object mode you have to go through all objects to clear sharp and bevel + the bevel modifier
        else:
            modifier_name = 'Bevel'
            
            # Loop through all objects
            for x in bpy.context.selected_objects:
                bpy.context.scene.objects.active = bpy.data.objects[x.name]
                
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='EDGE')
                
                # Clean slate. Remove previous bevel and sharp edges 
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.mark_sharp(clear=True)
                bpy.ops.transform.edge_bevelweight(value=-1)
                bpy.ops.mesh.select_all(action='DESELECT')
                
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                bpy.ops.object.modifier_remove(modifier = modifier_name)

        return {'FINISHED'}
        
        

# Add bevel modifier in object mode
class VIEW3D_OT_ko_add_bevel(bpy.types.Operator):
    """Tooltip"""
    bl_idname = "view3d.ko_add_bevel"
    bl_label = "kO Add Bevel"
    bl_options = {'REGISTER', 'UNDO'}
    
    loop_slide = BoolProperty(
        name="Loop Slide", 
        description="Prefer sliding along edges.", 
        default=True)
        
    clamp_overlap = BoolProperty(
        name="Clamp Overlap", 
        description="Clamp the width to avoid overlap.", 
        default=True)
        
    mark_sharp = BoolProperty(
        name="Mark Sharp", 
        description="Mark edges as sharp.", 
        default=True)
        
    angle_selection = FloatProperty(
        name="Angle Selection", 
        description="Select edges based on angle.", 
        default= 30,
        soft_min = 0,
        soft_max = 180)
    
    bevel_width = FloatProperty(
        name="Bevel Width", 
        description="Change the bevel width.", 
        default= 0.1,
        soft_min = 0,
        soft_max = 64)
        
    bevel_segments = IntProperty(
        name="Bevel Segments", 
        description="Change the bevel segments.", 
        default= 2,
        soft_min = 1,
        soft_max = 12)
        
    bevel_profile = FloatProperty(
        name="Bevel Profile", 
        description="Change the sharpness of the bevel profile.", 
        default= 0.5,
        soft_min = 0,
        soft_max = 1)
        
    do_once = True
    modifier_name = 'Bevel'
    modifier_created = False
    
    def setDefaults(self):
        if self.do_once:
            #bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
            for x in bpy.context.object.modifiers:
                if x.name == self.modifier_name: # check if bevel modifier exists already
                    # Store new defaults for the bevel
                    self.bevel_width = bpy.context.object.modifiers[self.modifier_name].width
                    self.bevel_segments = bpy.context.object.modifiers[self.modifier_name].segments
                    self.loop_slide = bpy.context.object.modifiers[self.modifier_name].loop_slide
                    self.bevel_profile = bpy.context.object.modifiers[self.modifier_name].profile
                    self.clamp_overlap = bpy.context.object.modifiers[self.modifier_name].use_clamp_overlap
                    self.do_once = False
                    print('Do Once')
                    break   
    
        return
    
    
    def createModifier(self):
    
        #bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
    
        for x in bpy.context.object.modifiers:
            
            if x.name == self.modifier_name: # check if bevel modifier exists already
                self.modifier_created = True 
                bpy.context.object.modifiers[self.modifier_name].show_viewport = True
                bpy.context.object.modifiers[self.modifier_name].show_in_editmode = True
                bpy.context.object.modifiers[self.modifier_name].show_on_cage = True
                    
                break           
            
        # Bevel modifier does not exist, so create one
        if self.modifier_created == False:
            bpy.ops.object.modifier_add(type='BEVEL')
            modifier_id = (len(bpy.context.object.modifiers)-1)
            bpy.context.object.modifiers[modifier_id].name = self.modifier_name
            bpy.context.object.modifiers[modifier_id].limit_method = 'WEIGHT'
            bpy.context.object.modifiers[modifier_id].offset_type = 'WIDTH'

        bpy.context.object.modifiers[self.modifier_name].width = self.bevel_width
        bpy.context.object.modifiers[self.modifier_name].segments = self.bevel_segments
        bpy.context.object.modifiers[self.modifier_name].loop_slide = self.loop_slide
        bpy.context.object.modifiers[self.modifier_name].profile = self.bevel_profile
        bpy.context.object.modifiers[self.modifier_name].use_clamp_overlap = self.clamp_overlap
        
        self.modifier_created = False
        return
    
    
    # Only work in the 3d view and object mode
    @classmethod
    def poll(cls, context):
        return context.area.type == 'VIEW_3D' and (bpy.context.active_object.mode == 'OBJECT' or bpy.context.active_object.mode == 'EDIT')

    def execute(self, context):
        
        mode = bpy.context.active_object.mode
        
        if bpy.context.active_object.mode == 'OBJECT':
            if len(bpy.context.selected_objects) == 0:
                self.report({'WARNING'}, 'Nothing is selected.')
                return {'CANCELLED'}

            # check if active object is part of the selection
            active_check = False
            for x in bpy.context.selected_objects:
                if x.name == bpy.context.scene.objects.active.name:
                    active_check = True
            if active_check == False:
                bpy.context.scene.objects.active = bpy.context.selected_objects[0]
    
        self.setDefaults()
        if mode == 'EDIT':
            #if self.do_once:
            self.createModifier()
            #bpy.ops.object.mode_set(mode='EDIT', toggle=False)
        
            if self.mark_sharp:
                bpy.ops.mesh.mark_sharp(clear=False)
            else:
                bpy.ops.mesh.mark_sharp(clear=True)
            bpy.ops.transform.edge_bevelweight(value=1)
            
        else:
            # Loop through all objects
            for x in bpy.context.selected_objects:
                bpy.context.scene.objects.active = bpy.data.objects[x.name]
        
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='EDGE')
                
                # Clean slate. Remove previous bevel and sharp edges 
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.mark_sharp(clear=True)
                bpy.ops.transform.edge_bevelweight(value=-1)
                bpy.ops.mesh.select_all(action='DESELECT')
                
                # Select based on sharpness and mark bevel and sharp edges
                bpy.ops.mesh.edges_select_sharp(sharpness = self.angle_selection / 57.2958)
                if self.mark_sharp:
                    bpy.ops.mesh.mark_sharp(clear=False)
                bpy.ops.transform.edge_bevelweight(value=1)
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                
                self.createModifier()
            
        return {'FINISHED'}

# Draws the Custom Menu
class VIEW3D_OT_knockoff_menu(bpy.types.Menu):
    bl_label = "KnockOff - Menu"
    bl_idname = "VIEW3D_OT_knockoff_menu"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_DEFAULT'
        layout.operator("view3d.ko_add_bevel")
        layout.operator("view3d.ko_clear_bevel")
        layout.operator("view3d.ko_edit_bevel")


# Calls the KnockOff Menu
class VIEW3D_OT_knockoff(bpy.types.Operator): 
    """Calls the KnockOff Menu"""
    bl_idname = "view3d.knockoff"
    bl_label = "KnockOff"
    
    @classmethod    
    def poll(cls, context):
        return context.area.type == 'VIEW_3D' and (bpy.context.active_object.mode == 'OBJECT' or bpy.context.active_object.mode == 'EDIT')

    def execute(self, context):
        bpy.ops.wm.call_menu(name=VIEW3D_OT_knockoff_menu.bl_idname)
        return {'FINISHED'}
        
        
def register():
    bpy.utils.register_class(VIEW3D_OT_knockoff)
    bpy.utils.register_class(VIEW3D_OT_knockoff_menu)
    bpy.utils.register_class(VIEW3D_OT_ko_edit_bevel)
    bpy.utils.register_class(VIEW3D_OT_ko_clear_bevel)
    bpy.utils.register_class(VIEW3D_OT_ko_add_bevel)


def unregister():
    bpy.utils.unregister_class(VIEW3D_OT_knockoff)
    bpy.utils.unregister_class(VIEW3D_OT_knockoff_menu)
    bpy.utils.unregister_class(VIEW3D_OT_ko_edit_bevel)
    bpy.utils.unregister_class(VIEW3D_OT_ko_clear_bevel)
    bpy.utils.unregister_class(VIEW3D_OT_ko_add_bevel)


if __name__ == "__main__":
    register()

    # test call
    #bpy.ops.mesh.ko_add_bevel()